-- Gwee: graphical user interface module for love2d
-- Author: Michael Carius <michael.a.carius@gmail.com>

module(..., package.seeall)

require "loo"

Group = loo.class {
    initialize = function(self, skin, layout)
        self.widgets = {}
        self.widgetArray = {}
        self.skin = skin or DEFAULT_SKIN
        self.layout = layout or NullLayout()
    end,

    addWidget = function(self, widget, id)
        self.widgets[id] = widget
        table.insert(self.widgetArray, widget)
        self:configure()
    end,

    configure = function(self)
        self.layout:configure(self.widgetArray)
    end,

    keypressed = function(self, key, unicode)
        for id, widget in pairs(self.widgets) do
            if widget.keypressed then
                widget:keypressed(key, unicode)
            end
        end
    end,

    mousepressed = function(self, x, y, button)
        for id, widget in pairs(self.widgets) do
            if widget.mousepressed then
                widget:mousepressed(x, y, button)
            end
        end
    end,

    mousereleased = function(self, x, y, button)
        for id, widget in pairs(self.widgets) do
            if widget.mousereleased then
                widget:mousereleased(x, y, button)
            end
        end
    end,

    update = function(self, dt)
        for id, widget in pairs(self.widgets) do
            if widget.update then
                widget:update(dt)
            end
        end
    end,

    draw = function(self)
        for id, widget in pairs(self.widgets) do
            if widget.drawWithSkin then
                widget:drawWithSkin(self.skin)
            end
        end
    end,
}

Button = loo.class {
    initialize = function(self, label, callback, box)
        self.label = label
        self.callback = callback
        self.box = box or Box(0, 0, 0, 0)
    end,

    keypressed = function(self, key, unicode)
        if key == "return" and self.box.focused then
            self.callback()
        end
    end,

    mousepressed = function(self, x, y, button)
        self.box:mousepressed(x, y, button)
        if self.box.focused then
            self.callback()
        end
    end,

    update = function(self, dt)
        self.box:update(dt)
    end,

    drawWithSkin = function(self, skin)
        local style = skin.button

        -- Draw the box
        if self.box.hover then
            love.graphics.setColor(style.hovercolor)
        else
            love.graphics.setColor(style.color)
        end
        love.graphics.rectangle("fill", self.box.x, self.box.y, self.box.w, self.box.h)

        -- Draw the border
        -- 2px hack
        love.graphics.setLineWidth(2)
        love.graphics.setColor(style.bordercolor)
        love.graphics.rectangle("line", self.box.x, self.box.y, self.box.w, self.box.h)

        -- Draw the label
        local labelX = self.box.x + (self.box.w - style.font:getWidth(self.label)) / 2
        local labelY = self.box.y + (self.box.h - style.font:getHeight()) / 2
        love.graphics.setColor(style.labelcolor)
        love.graphics.setFont(style.font)
        love.graphics.print(self.label, labelX, labelY)
    end,
}

TextField = loo.class {
    initialize = function(self, label, box)
        self.label = label
        self.box = box or Box(0, 0, 0, 0)
        self.text = ""
        self.pos = 0
    end,

    keypressed = function(self, key, unicode)
        if self.box.focused then
            if key == "backspace" then
                self:backspace()
            elseif key == "delete" then
                self:delete()
            elseif key == "home" then
                self:goToBeginning()
            elseif key == "end" then
                self:goToEnd()
            elseif key == "left" then
                self:left()
            elseif key == "right" then
                self:right()
            -- assume char length of 1 is a printable char
            elseif key:len() == 1 then
                self:insertCharacter(string.char(unicode))
            end
        end
    end,

    insertCharacter = function(self, character)
        self.text = self.text:sub(1, self.pos) .. character .. self.text:sub(self.pos + 1)
        self.pos = self.pos + 1
    end,

    backspace = function(self)
        if self.pos > 0 then
            self.text = self.text:sub(1, self.pos - 1) .. self.text:sub(self.pos + 1)
            self.pos = self.pos - 1
        end
    end,

    delete = function(self)
        if self.pos < self.text:len() then
            self.text = self.text:sub(1, self.pos) .. self.text:sub(self.pos + 2)
        end
    end,

    goToEnd = function(self)
        self.pos = self.text:len()
    end,

    goToBeginning = function(self)
        self.pos = 0
    end,

    left = function(self)
        if self.pos > 0 then
            self.pos = self.pos - 1
        end
    end,

    right = function(self)
        if self.pos < self.text:len() then
            self.pos = self.pos + 1
        end
    end,

    mousepressed = function(self, x, y, button)
        self.box:mousepressed(x, y, button)
    end,

    drawWithSkin = function(self, skin)
        local style = skin.textfield

        -- Draw the label
        local labelX = self.box.x - style.font:getWidth(self.label) - style.labelgap
        local labelY = self.box.y + (self.box.h - style.font:getHeight()) / 2
        love.graphics.setColor(style.labelcolor)
        love.graphics.setFont(style.font)
        love.graphics.print(self.label, labelX, labelY)

        -- Draw the box
        love.graphics.setColor(style.color)
        love.graphics.rectangle("fill", self.box.x, self.box.y, self.box.w, self.box.h)

        -- Draw the border
        -- 2px hack
        love.graphics.setLineWidth(2)
        love.graphics.setColor(style.bordercolor)
        love.graphics.rectangle("line", self.box.x, self.box.y, self.box.w, self.box.h)

        -- Draw the text
        -- Scissoring means no fancy clipping logic
        -- (clips to a box smaller than the textbox by 2px on either side)
        love.graphics.setScissor(self.box.x + style.margins.left, self.box.y, self.box.w - (style.margins.left + style.margins.right), self.box.h)

        local textWidth = style.font:getWidth(self.text:sub(1, self.pos))
        local textX
        if textWidth > self.box.w then
            -- Right-justify to pos if text is longer than the box
            textX = (self.box.x + self.box.w) - (textWidth - self.box.w) - self.box.w - style.margins.right
        else
            -- Left-justify otherwise
            textX = self.box.x + style.margins.left
        end
        -- Text is at the same level as the label
        local textY = labelY
        love.graphics.setColor(style.textcolor)
        love.graphics.setFont(style.font)
        love.graphics.print(self.text, textX, textY)

        -- Disable scissoring
        love.graphics.setScissor()

        -- Draw the cursor
        if self.box.focused then
            local cursorHeight = self.box.h * style.cursorheight
            local cursorVertices = {
                textX + textWidth, self.box.y + (self.box.h - cursorHeight) / 2,
                textX + textWidth, self.box.y + cursorHeight
            }

            -- 2px hack
            love.graphics.setLineWidth(2)
            love.graphics.setColor(style.textcolor)
            love.graphics.line(cursorVertices)
        end
    end,
}

CheckBox = loo.class {
    initialize = function(self, label, box)
        self.label = label
        self.box = box or Box(0, 0, 0, 0)
        self.checked = false
    end,

    keypressed = function(self, key, unicode)
        if key == " " and self.box.focused then
            self.checked = not self.checked
        end
    end,

    mousepressed = function(self, x, y, button)
        self.box:mousepressed(x, y, button)
        if self.box.focused then
            self.checked = not self.checked
        end
    end,

    drawWithSkin = function(self, skin)
        local style = skin.checkbox

        -- Hack -- it's a *box*, so we adjust to the biggest square we can get
        -- away with
        local boxX, boxY, boxSize
        if self.box.w < self.box.h then
            boxSize = self.box.w
            boxX = self.box.x
            boxY = self.box.y + (self.box.h - boxSize) / 2
        else
            boxSize = self.box.h
            boxX = self.box.x
            boxY = self.box.y
        end

        -- Draw the label
        local labelX = self.box.x - style.font:getWidth(self.label) - 5
        local labelY = self.box.y + (self.box.h - style.font:getHeight()) / 2
        love.graphics.setColor(style.labelcolor)
        love.graphics.setFont(style.font)
        love.graphics.print(self.label, labelX, labelY)

        -- Draw the box
        love.graphics.setColor(style.color)
        love.graphics.rectangle("fill", boxX, boxY, boxSize, boxSize)

        -- Draw the border
        -- 2px hack
        love.graphics.setLineWidth(2)
        love.graphics.setColor(style.bordercolor)
        love.graphics.rectangle("line", boxX, boxY, boxSize, boxSize)

        -- Draw the mark
        if self.checked then
            love.graphics.setColor(style.markcolor)
            if style.markstyle == "circle" then
                local markCenterX = boxX + boxSize / 2
                local markCenterY = boxY + boxSize / 2
                local markRadius = (boxSize / 2) * style.marksize
                local numSegments = 20
                love.graphics.circle("fill", markCenterX, markCenterY, markRadius, numSegments)
            elseif style.markstyle == "square" then
                local markSize = boxSize * style.marksize
                local markX = boxX + (boxSize - markSize) / 2
                local markY = boxY + (boxSize - markSize) / 2
                love.graphics.rectangle("fill", markX, markY, markSize, markSize)
            end
        end
    end,
}

Slider = loo.class {
    initialize = function(self, min, max, label, box)
        self.min = min
        self.range = max - min
        self.label = label
        self.box = box or Box(0, 0, 0, 0)
        self.value = min + self.range / 2
    end,

    keypressed = function(self, key, unicode)
        if self.box.focused then
            if key == "left" then
                self:less()
            elseif key == "right" then
                self:more()
            end
        end
    end,

    addValue = function(self, amount)
        self.value = self:clamp(self.value + amount)
    end,

    clamp = function(self, value)
        if value > self.min + self.range then
            return self.min + self.range
        elseif value < self.min then
            return self.min
        else
            return value
        end
    end,

    less = function(self)
        self:addValue(-self.range / 10)
    end,

    more = function(self)
        self:addValue(self.range / 10)
    end,

    mousepressed = function(self, x, y, button)
        self.box:mousepressed(x, y, button)
        if self.box.focused then
            self.value = self:valueForMousePosition(x)
        end
    end,

    mousereleased = function(self, x, y, button)
        if self.box.dragging then
            self.value = self:valueForMousePosition(x)
        end
        self.box:mousereleased(x, y, button)
    end,

    update = function(self, dt)
        if self.box.dragging then
            self.value = self:valueForMousePosition(love.mouse.getX())
        end
    end,

    valueForMousePosition = function(self, mouseX)
        local percent = (mouseX - self.box.x) / self.box.w
        return self:clamp(self.min + self.range * percent)
    end,

    drawWithSkin = function(self, skin)
        local style = skin.slider

        -- Draw the label
        local labelX = self.box.x - style.font:getWidth(self.label) - 5
        local labelY = self.box.y + (self.box.h - style.font:getHeight()) / 2
        love.graphics.setColor(style.labelcolor)
        love.graphics.setFont(style.font)
        love.graphics.print(self.label, labelX, labelY)

        -- Draw the bar
        local barHeight = self.box.h * style.barsize
        local barY = self.box.y + (self.box.h - barHeight) / 2
        love.graphics.setColor(style.barcolor)
        love.graphics.rectangle("fill", self.box.x, barY, self.box.w, barHeight)

        -- Draw the slider
        local percent = (self.value - self.min) / self.range
        local sliderX = self.box.x + self.box.w * percent
        local sliderWidth = self.box.w * style.sliderwidth
        love.graphics.setColor(style.slidercolor)
        love.graphics.rectangle("fill", sliderX, self.box.y, sliderWidth, self.box.h)
    end,
}

Text = loo.class {
    initialize = function(self, text, justify, box)
        self.text = text
        self.justify = justify or "center"
        self.box = box or Box(0, 0, 0, 0)
    end,

    drawWithSkin = function(self, skin)
        local style = skin.text
        local textX, textY
        if self.justify == "center" then
            textX = self.box.x + (self.box.w - style.font:getWidth(self.text)) / 2
        elseif self.justify == "right" then
            textX = self.box.x + self.box.w - style.font:getWidth(self.text)
        elseif self.justify == "left" then
            textX = self.box.x
        end
        textY = self.box.y + (self.box.h - style.font:getHeight()) / 2
        love.graphics.setFont(style.font)
        love.graphics.setColor(style.color)
        love.graphics.print(self.text, textX, textY)
    end,
}

Box = loo.class {
    initialize = function(self, x, y, w, h)
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.hover = false
        self.focused = false
        self.dragging = false
    end,

    mousepressed = function(self, x, y, button)
        if self:contains(x, y) then
            self.focused = true
            self.dragging = true
        else
            self.focused = false
        end
    end,

    mousereleased = function(self, x, y, button)
        if self.dragging then
            self.dragging = false
        end
    end,

    update = function(self, dt)
        local x, y = love.mouse.getPosition()
        if self:contains(x, y) then
            self.hover = true
        else
            self.hover = false
        end
    end,

    contains = function(self, x, y)
        local collideX = x > self.x and x < self.x + self.w
        local collideY = y > self.y and y < self.y + self.h
        return collideX and collideY
    end,
}

NullLayout = loo.class {
    configure = function(self, widgets)
    end
}

HorizontalLayout = loo.class {
    initialize = function(self, box, spacing)
        self.box = box
        self.spacing = spacing or 0
    end,

    configure = function(self, widgets)
        local widgetWidth = (self.box.w - self.spacing * (#widgets + 1)) / #widgets
        local widgetHeight = self.box.h - self.spacing * 2
        local widgetX = self.box.x + self.spacing
        local widgetY = self.box.y + self.spacing
        for _, widget in ipairs(widgets) do
            widget.box = Box(widgetX, widgetY, widgetWidth, widgetHeight)
            widgetX = widgetX + self.spacing + widgetWidth
        end
    end,
}

VerticalLayout = loo.class {
    initialize = function(self, box, spacing)
        self.box = box
        self.spacing = spacing or 0
    end,

    configure = function(self, widgets)
        local widgetWidth = self.box.w - self.spacing * 2
        local widgetHeight = (self.box.h - self.spacing * (#widgets + 1)) / #widgets
        local widgetX = self.box.x + self.spacing
        local widgetY = self.box.y + self.spacing
        for _, widget in ipairs(widgets) do
            widget.box = Box(widgetX, widgetY, widgetWidth, widgetHeight)
            widgetY = widgetY + self.spacing + widgetHeight
        end
    end,
}

function loadSkin(skinName)
    -- load the skin itself
    local skin = love.filesystem.load(skinName .. "/style.lua")()

    -- load resources specified in the skin
    skin.button.font = skin.button.font and love.graphics.newFont(skinName .. "/" .. skin.button.font, skin.button.fontsize)
    skin.textfield.font = skin.textfield.font and love.graphics.newFont(skinName .. "/" .. skin.textfield.font, skin.textfield.fontsize)
    skin.checkbox.font = skin.checkbox.font and love.graphics.newFont(skinName .. "/" .. skin.checkbox.font, skin.checkbox.fontsize)
    skin.slider.font = skin.slider.font and love.graphics.newFont(skinName .. "/" .. skin.slider.font, skin.slider.fontsize)
    skin.text.font = skin.text.font and love.graphics.newFont(skinName .. "/" .. skin.text.font, skin.text.fontsize)

    -- set defaults
    for widgetClass, style in pairs(DEFAULT_SKIN) do
        for k, defaultValue in pairs(style) do
            if skin[widgetClass][k] == nil then
                skin[widgetClass][k] = defaultValue
            end
        end
    end

    return skin
end

DEFAULT_SKIN = {
    button = {
        font = love.graphics.newFont(12),
        fontsize = 12,
        labelcolor = {0x00, 0x00, 0x00, 0xff},
        color = {0x44, 0x44, 0x44, 0xff},
        hovercolor = {0x88, 0x88, 0x88, 0xff},
        bordercolor = {0x66, 0x66, 0x66, 0xff}
    },

    textfield = {
        font = love.graphics.newFont(12),
        fontsize = 12,
        labelcolor = {0xff, 0xff, 0xff, 0xff},
        labelgap = 5,
        textcolor = {0x00, 0x00, 0x00, 0xff},
        margins = {left = 2, right = 2},
        cursorheight = 0.9,
        color = {0xff, 0xff, 0xff, 0xff},
        bordercolor = {0xaa, 0xaa, 0xaa, 0xff}
    },

    checkbox = {
        font = love.graphics.newFont(12),
        fontsize = 12,
        labelcolor = {0xff, 0xff, 0xff, 0xff},
        color = {0xff, 0xff, 0xff, 0xff},
        bordercolor = {0xaa, 0xaa, 0xaa, 0xff},
        markcolor = {0x00, 0x00, 0x00, 0xff},
        markstyle = "circle",
        marksize = 0.6
    },

    slider = {
        font = love.graphics.newFont(12),
        fontsize = 12,
        labelcolor = {0xff, 0xff, 0xff, 0xff},
        barcolor = {0x44, 0x44, 0x44, 0x44, 0xff},
        barsize = 0.5,
        slidercolor = {0x88, 0x88, 0x88, 0xff},
        sliderwidth = 0.1,
    },

    text = {
        font = love.graphics.newFont(12),
        fontsize = 12,
        color = {0xff, 0xff, 0xff, 0xff}
    }
}
