-- B concept
-- This implementation may or may not be used to make the final game
-- Author: Michael Carius <michael.a.carius@gmail.com>

require "loo"
require "sat"
require "gwee"

Camera = loo.class {
    initialize = function(self)
        self.x = 0
        self.y = 0
    end,

    clamp = function(self, x, y, w, h)
        if self.x < x then
            self.x = x
        elseif self.x + love.graphics.getWidth() > x + w then
            self.x = x + w - love.graphics.getWidth()
        end

        if self.y < y then
            self.y = y
        elseif self.y + love.graphics.getHeight() > y + h then
            self.y = y + h - love.graphics.getHeight()
        end
    end,

    follow = function(self, object)
        self.follow = object
    end,

    update = function(self, dt)
        if self.follow then
            -- center on the followed object
            self.x = self.follow.shape.center.x - love.graphics.getWidth() / 2
            self.y = self.follow.shape.center.y - love.graphics.getHeight() / 2
        end
    end,

    set = function(self)
        love.graphics.push()
        love.graphics.translate(-self.x, -self.y)
    end,

    unset = function(self)
        love.graphics.pop()
    end,
}

MagneticObject = loo.class {
    initialize = function(self, shape, q)
        table.insert(shapes, shape)
        self.shape = shape
        self.q = q
    end,

    draw = function(self)
        if self.q < 0 then
            -- South
            love.graphics.setColor(0x00, 0x00, 0xff, 0xff)
        elseif self.q > 0 then
            -- North
            love.graphics.setColor(0xff, 0x00, 0x00, 0xff)
        else
            -- Neutral/not magnetic
            love.graphics.setColor(0x00, 0xff, 0x00, 0xff)
        end
        love.graphics.circle("fill", self.shape.center.x, self.shape.center.y, self.shape.radius, 20)
    end,
}

OrbitingObject = loo.class {
    initialize = function(self, centerX, centerY, orbitRadius, orbitSpeed, object)
        self.centerX = centerX
        self.centerY = centerY
        self.orbitRadius = orbitRadius
        self.orbitSpeed = orbitSpeed
        self.object = object
        object.shape.center.x = centerX + orbitRadius
        object.shape.center.y = centerY
        self.angle = 0
    end,

    update = function(self, dt)
        self.angle = self.angle + self.orbitSpeed * dt
        self.object.shape.center.x = self.centerX + self.orbitRadius * math.cos(self.angle)
        self.object.shape.center.y = self.centerY + self.orbitRadius * -math.sin(self.angle)
    end,
}

Surface = loo.class {
    initialize = function(self, shape)
        table.insert(shapes, shape)
        self.shape = shape
    end,

    draw = function(self)
        love.graphics.setColor(0x44, 0x44, 0x44, 0xff)
        love.graphics.polygon("fill", unpack(self.shape:getCorners()))
        love.graphics.setLineWidth(2)
        love.graphics.setColor(0x88, 0x88, 0x88, 0xff)
        love.graphics.polygon("line", unpack(self.shape:getCorners()))
    end,
}

-- Stub
Goal = loo.class {
    initialize = function(self, x, y)
        self.x = x
        self.y = y
    end,

    draw = function(self)
        love.graphics.setFont(goalFont)
        love.graphics.setColor(0xff, 0xff, 0x00, 0xff)
        love.graphics.print("Goal", self.x, self.y)
    end,
}

-- Stub
DeathPit = loo.class {
    initialize = function(self, x, y, w)
        self.x = x
        self.y = y
        self.w = w
    end,

    draw = function(self)
        love.graphics.setFont(deathPitFont)
        love.graphics.setColor(0xff, 0x66, 0x00, 0xff)
        love.graphics.printf("Death Pit (if you fall in here you are \"dead\")", self.x, self.y, self.w, "center")
    end,
}

function collideCB(a, b, mtv)
    local s
    if a.id == "player" then
        s = b
    elseif b.id == "player" then
        s = a
    else
        return
    end

    if s.id == "magnet" then
        -- velocity
        -- Subtract the component of velocity that is pointed toward the
        -- center of m
        local v = math.sqrt(player.dx ^ 2 + player.dy ^ 2)
        local cos = (player.shape.center.x - s.center.x) / sat.distance(player.shape, s)
        local sin = (player.shape.center.y - s.center.y) / sat.distance(player.shape, s)
        local p = player.dx * cos + player.dy * sin
        player.dx = player.dx - p * cos
        player.dy = player.dy - p * sin

        -- position
        -- mtv is backwards along penetration vector
        local d = player.shape.radius + s.radius - sat.distance(player.shape, s)
        local cos = (player.shape.center.x - s.center.x) / sat.distance(player.shape, s)
        local sin = (player.shape.center.y - s.center.y) / sat.distance(player.shape, s)
        local dx = cos * d
        local dy = sin * d
        player.shape.center.x = player.shape.center.x + dx
        player.shape.center.y = player.shape.center.y + dy
    elseif s.id == "surface" then
        -- velocity
        -- Same as above
        local v = math.sqrt(player.dx ^ 2 + player.dy ^ 2)
        local n = mtv:normalized()
        local dp = sat.dotProduct(sat.Vector(player.dx, player.dy), n)
        player.dx = player.dx - dp * n.x
        player.dy = player.dy - dp * n.y

        -- flags
        player.jumping = false
        player.friction = true
        player.airborn = false

        -- position
        if player.shape.center.x < s.center.x then
            player.shape.center.x = player.shape.center.x - mtv.x
        else
            player.shape.center.x = player.shape.center.x + mtv.x
        end
        if player.shape.center.y < s.center.y then
            player.shape.center.y = player.shape.center.y - mtv.y
        else
            player.shape.center.y = player.shape.center.y + mtv.y
        end
    end
end

shapes = {}

magnets = {}
function makeMagnet(x, y, q)
    local m = MagneticObject(sat.Circle("magnet", x, y, math.abs(q)), q)
    table.insert(magnets, m)
end

orbits = {}
function makeOrbitingMagnet(cx, cy, r, w, q)
    makeMagnet(0, 0, q)
    table.insert(orbits, OrbitingObject(cx, cy, r, w, magnets[#magnets]))
end

surfaces = {}
function makeSurface(x, y, w, h)
    table.insert(surfaces, Surface(sat.Rectangle("surface", x, y, w, h)))
end

function love.load()
    LEVEL_WIDTH = 2200
    LEVEL_HEIGHT = 1250

    goalFont = love.graphics.newFont(18)
    deathPitFont = love.graphics.newFont(18)

    player = MagneticObject(sat.Circle("player", 30, LEVEL_HEIGHT - 100, 10), 0)
    player.impulses = {x = 0, y = 0}
    player.dx = 0
    player.dy = 0

    --[[
    -- Small test level
    makeSurface(0, love.graphics.getHeight() - 30, love.graphics.getWidth(), 30)

    makeMagnet(150, 500, 10)
    makeMagnet(650, 500, 10)
    makeMagnet(170, 450, 10)
    makeMagnet(630, 450, 10)
    makeMagnet(190, 400, 10)
    makeMagnet(610, 400, 10)
    makeMagnet(210, 350, 10)
    makeMagnet(590, 350, 10)
    makeMagnet(250, 300, 20)
    makeMagnet(550, 300, 20)
    makeMagnet(385, 550, -10)
    makeMagnet(415, 550, -10)
    makeSurface(385, 400, 30, 30)

    makeMagnet(400, 300, 30)
    makeOrbitingMagnet(400, 300, 40, .5, 10)
    makeOrbitingMagnet(400, 300, 60, 2, -5)
    ]]

    -- The demo level!
    makeSurface(0, 1200, 400, 50)
    makeSurface(400, 900, 600, 350)
    makeSurface(400, 0, 50, 800)
    makeSurface(1400, 900, 800, 350)
    makeSurface(600, 300, 400, 100)

    makeMagnet(200, 820, 40)
    makeOrbitingMagnet(1200, 800, 100, 2, 20)
    makeMagnet(1985, 900, -10)
    makeMagnet(2015, 900, -10)
    makeMagnet(1900, 845, 10)
    makeMagnet(2100, 845, 10)
    makeMagnet(1900, 795, 10)
    makeMagnet(2100, 795, 10)
    makeMagnet(1900, 745, 10)
    makeMagnet(2100, 745, 10)
    makeMagnet(1900, 695, 10)
    makeMagnet(2100, 695, 10)
    makeMagnet(1900, 645, 10)
    makeMagnet(2100, 645, 10)
    makeMagnet(1900, 595, 10)
    makeMagnet(2100, 595, 10)
    makeMagnet(1900, 545, 10)
    makeMagnet(2100, 545, 10)
    makeMagnet(2000, 50, 10)
    makeMagnet(1500, 250, 40)
    makeMagnet(600, 50, 5)
    makeMagnet(600, 100, 5)
    makeMagnet(600, 150, 5)
    makeMagnet(600, 200, 5)
    makeMagnet(600, 250, 5)

    goal = Goal(700, 250)
    deathPit = DeathPit(1000, 1150, 400)

    -- forces
    gravity = 10
    friction = 10

    cam = Camera()
    cam:follow(player)

    state = "menu"
    local guiSkin = gwee.loadSkin("gui-skin")
    header = gwee.Group(guiSkin)
    header:addWidget(gwee.Text("B Concept Demo", "center", gwee.Box(350, 50, 100, 100)), "1")
    menu = gwee.Group(guiSkin)
    menu.layout = gwee.VerticalLayout(gwee.Box(350, 250, 100, 100), 20)
    menu:addWidget(gwee.Button("Play", function() state = "game" end), "play")
    menu:addWidget(gwee.Button("Quit", function() love.event.push("q") end), "quit")
    footer = gwee.Group(guiSkin)
    footer.layout = gwee.VerticalLayout(gwee.Box(350, 450, 100, 100), 10)
    footer:addWidget(gwee.Text("Built with love2d (http://love2d.org/)"), "1")
    footer:addWidget(gwee.Text("All programming and game design by Michael Carius"), "2")
end

function love.mousepressed(x, y, button)
    if state == "menu" then
        menu:mousepressed(x, y, button)
    end
end

function love.mousereleased(x, y, button)
    if state == "menu" then
        menu:mousereleased(x, y, button)
    end
end

function love.keypressed(key, unicode)
    if state == "menu" then
        menu:keypressed(key, unicode)
    else
        if key == "escape" then
            state = "menu"
            return
        end
        local impulse = 0
        if not player.airborn then
            if key == "a" then
                --player.dx = player.dx - 500
                impulse = -200
            elseif key == "d" then
                --player.dx = player.dx + 500
                impulse = 200
            end
        else
            if key == "a" then
                impulse = -50
            elseif key == "d" then
                impulse = 50
            end
        end
        player.dx = player.dx + impulse
        player.impulses.x = player.impulses.x + impulse
        if key == "h" then
            if not player.jumping then
                player.jumping = true
                player.dy = player.dy - 300
            end
        elseif key == "k" then
            player.q = player.q + 100
        elseif key == "l" then
            player.q = player.q - 100
        end
    end
end

function love.keyreleased(key, unicode)
    if state == "menu" then
        return
    end
    if key == "a" then
        player.dx = math.min(player.dx + 200, 0)
        player.impulses.x = math.min(player.impulses.x + 200, 0)
    elseif key == "d" then
        player.dx = math.max(player.dx - 200, 0)
        player.impulses.x = math.max(player.impulses.x - 200, 0)
    elseif key == "k" then
        player.q = player.q - 100
    elseif key == "l" then
        player.q = player.q + 100
    end
end

function love.update(dt)
    if state == "menu" then
        menu:update(dt)
        return
    end
    -- Update normal stuff
    player.dy = player.dy + gravity
    player.shape.center.x = player.shape.center.x + player.dx * dt
    player.shape.center.y = player.shape.center.y + player.dy * dt

    for _, o in ipairs(orbits) do
        o:update(dt)
    end

    player.friction = false
    player.airborn = true
    sat.sat(shapes, collideCB)

    -- friction
    if player.friction then
        if player.dx > 0 then
            player.dx = math.max(player.dx - friction, player.impulses.x)
        else
            player.dx = math.min(player.dx + friction, player.impulses.x)
        end
    end

    for _, m in ipairs(magnets) do
        -- MAGNET STUFF
        -- Only magnets within (screen size / 2) affect the player
        local dx = math.abs(player.shape.center.x - m.shape.center.x)
        local dy = math.abs(player.shape.center.y - m.shape.center.y)
        if dx < love.graphics.getWidth() / 2 and dy < love.graphics.getHeight() / 2 then
            local r = sat.distance(player.shape, m.shape)
            local cos = (player.shape.center.x - m.shape.center.x) / r
            local sin = (player.shape.center.y - m.shape.center.y) / r
            local f = player.q * m.q / r
            player.dx = player.dx + f * cos
            player.dy = player.dy + f * sin
        end
    end

    -- screen boundary checks
    if player.shape.center.y - player.shape.radius < 0 then
        player.shape.center.y = 0 + player.shape.radius
        player.dy = 0
    elseif player.shape.center.y + player.shape.radius > LEVEL_HEIGHT then
        player.shape.center.y = LEVEL_HEIGHT - player.shape.radius
        player.dy = 0
    end
    if player.shape.center.x - player.shape.radius < 0 then
        player.shape.center.x = 0 + player.shape.radius
        player.dx = 0
    elseif player.shape.center.x + player.shape.radius > LEVEL_WIDTH then
        player.shape.center.x = LEVEL_WIDTH - player.shape.radius
        player.dx = 0
    end

    cam:update(dt)
    cam:clamp(0, 0, LEVEL_WIDTH, LEVEL_HEIGHT)
end

function love.draw()
    if state == "menu" then
        header:draw()
        menu:draw()
        footer:draw()
        return
    end
    cam:set()
    player:draw()

    for _, m in ipairs(magnets) do
        m:draw()
    end
    for _, s in ipairs(surfaces) do
        s:draw()
    end
    goal:draw()
    deathPit:draw()
    cam:unset()
end
