-- Collision detection framework using the Separating Axis Theorem
-- Some code copied from HardonCollider
-- (http://vrld.github.com/HardonCollider/index.html)
-- Author: Michael Carius <michael.a.carius@gmail.com>

module(..., package.seeall)

require "loo"

Circle = loo.class {
    initialize = function(self, id, centerX, centerY, radius)
        self.id = id
        self.center = {x = centerX, y = centerY}
        self.radius = radius
    end,

    getAxesRelativeTo = function(self, relativeTo)
        -- Axis is from our center to the other shape's center
        local d = distance(self, relativeTo)
        local axisX = (self.center.x - relativeTo.center.x)
        local axisY = (self.center.y - relativeTo.center.y)
        return {Vector(axisX, axisY):normalized()}
    end,

    project = function(self, axis)
        local centerDisplacement = Vector(self.center.x, self.center.y)
        local projectedCenter = dotProduct(centerDisplacement:projected(axis), axis)
        local shift = self.radius * dotProduct(axis, axis)
        return {left = projectedCenter - shift, right = projectedCenter + shift}
    end,
}

Rectangle = loo.class {
    initialize = function(self, id, x, y, width, height)
        self.id = id
        self.center = {x = x + width / 2, y = y + height / 2}
        self.corners = {
            {x = x, y = y},
            {x = x + width, y = y},
            {x = x + width, y = y + height},
            {x = x, y = y + height}
        }
    end,

    getAxesRelativeTo = function(self, relativeTo)
        -- Axes are normals
        return {Vector(1, 0), Vector(0, 1)}
    end,

    project = function(self, axis)
        local projectedCorners = {}
        for _, corner in ipairs(self.corners) do
            local displacement = Vector(corner.x, corner.y)
            table.insert(projectedCorners, dotProduct(displacement:projected(axis), axis))
        end
        return {left = math.min(unpack(projectedCorners)), right = math.max(unpack(projectedCorners))}
    end,

    getCorners = function(self)
        local stuff = {}
        for _, v in ipairs(self.corners) do
            table.insert(stuff, v.x)
            table.insert(stuff, v.y)
        end

        return stuff
    end,
}

Vector = loo.class {
    initialize = function(self, x, y)
        self.x = x
        self.y = y
    end,

    __sub = function(self, rhs)
        return Vector(self.x - rhs.x, self.y - rhs.y)
    end,

    __eq = function(self, rhs)
        return self.x == rhs.x and self.y == rhs.y
    end,

    __lt = function(self, rhs)
        return self ~= rhs and self.x <= rhs.x and self.y <= rhs.y
    end,

    length = function(self)
        return math.sqrt(self.x ^ 2 + self.y ^ 2)
    end,

    __tostring = function(self)
        return "(" .. self.x .. ", " .. self.y .. ")"
    end,

    normalized = function(self)
        local d = self:length()
        if d == 0 then
            return Vector(0, 0)
        else
            return Vector(self.x / d, self.y / d)
        end
    end,

    normals = function(self)
        local left = Vector(-self.y, self.x)
        local right = Vector(self.y, -self.x)
        return left, right
    end,

    projected = function(self, onto)
        local dp = dotProduct(self, onto)
        local s = dotProduct(onto, onto)
        return Vector((dp / s) * onto.x, (dp / s) * onto.y)
    end,
}

function sat(shapes, callback)
    for _, shapeA in ipairs(shapes) do
        for _, shapeB in ipairs(shapes) do
            testShapes(shapeA, shapeB, callback)
        end
    end
end

function testShapes(shapeA, shapeB, callback)
    if shapeA ~= shapeB then
        local axes = getAllAxes(shapeA, shapeB)
        local collide, mtv = testAxes(axes, shapeA, shapeB)
        if collide then
            callback(shapeA, shapeB, mtv)
        end
    end
end

function getAllAxes(shapeA, shapeB)
    local axes = {}
    
    for _, v in ipairs(shapeA:getAxesRelativeTo(shapeB)) do
        table.insert(axes, v)
    end
    for _, v in ipairs(shapeB:getAxesRelativeTo(shapeA)) do
        table.insert(axes, v)
    end

    return axes
end

function testAxes(axes, shapeA, shapeB)
    local separated = false
    local i = 1
    local mtvSoFar
    local mtv
    local mtvAxis
    while i <= #axes and not separated do
        local axis = axes[i]
        i = i + 1
        local projectionA = shapeA:project(axis)
        local projectionB = shapeB:project(axis)
        separated, mtv = testProjections(projectionA, projectionB)
        if not mtvSoFar or mtv < mtvSoFar then
            mtvSoFar = mtv
            mtvAxis = axis
        end
    end
    return not separated, Vector(mtvSoFar * mtvAxis.x, mtvSoFar * mtvAxis.y)
end

function testProjections(projectionA, projectionB)
    local notSeparated = projectionA.right < projectionB.left or projectionA.left > projectionB.right
    local absLeft = math.abs(projectionA.right - projectionB.left)
    local absRight = math.abs(projectionB.right - projectionA.left)
    local smallestOverlap = absLeft < absRight and (projectionA.right - projectionB.left) or (projectionB.right - projectionA.left)
    return notSeparated, smallestOverlap
end

function dotProduct(vectorA, vectorB)
    return vectorA.x * vectorB.x + vectorA.y * vectorB.y
end

function distance(shapeA, shapeB)
    local dx = shapeA.center.x - shapeB.center.x
    local dy = shapeA.center.y - shapeB.center.y
    return math.sqrt(dx ^ 2 + dy ^ 2)
end
